/*----------------------------------------------------------------------------

   libtunepimp -- The MusicBrainz tagging library.  
                  Let a thousand taggers bloom!
   
   Copyright (C) Robert Kaye 2003
   
   This file is part of libtunepimp.

   libtunepimp is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   libtunepimp is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with libtunepimp; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   $Id: mp3.cpp,v 1.6 2004/04/27 21:11:20 robert Exp $

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#ifdef WIN32
#include <winsock.h>
#else
#include <netinet/in.h>
#include <sys/param.h>
#endif

#include "plugin.h"
#include "mp3decode.h"
#include "metadata.h"
#include "id3_meta.h"

extern "C"
{

/*-------------------------------------------------------------------------*/

#ifndef WIN32
  #define initPlugin mp3InitPlugin
#endif

Plugin                  *initPlugin      (void);
static void              MP3Shutdown     (void);
static const char       *MP3GetVersion   (void);
static const char       *MP3GetName      (void);
static int               MP3GetNumFormats(void);
static int               MP3GetFormat    (int, char ext[TP_EXTENSION_LEN],
                                          char desc[TP_PLUGIN_DESC_LEN], int *functions);
static const char       *MP3GetError     (void);
static int               MP3WriteMetadata(const metadata_t *mdata, const char *fileName, int flags);
static int               MP3ReadMetadata (metadata_t *mdata, const char *fileName, int flags);
static void             *MP3DecodeStart  (const char *file);
static int               MP3DecodeInfo   (void *decode,
                                          unsigned long *duration,
                                          unsigned int *samplesPerSecond,
                                          unsigned int *bitsPerSample,
                                          unsigned int *channels);
static int               MP3DecodeRead   (void *decode, char *data, int maxBytes);
static void              MP3DecodeEnd    (void *decode);

/*-------------------------------------------------------------------------*/

#define PLUGIN_VERSION               "1.0.0"
#define PLUGIN_NAME                  "MP3 decoder & ID3v1/v2 reader/writer"
#define PLUGIN_FLAGS_WRITE_ID3V1     TP_PLUGIN_FLAGS_SPECIFIC1 

static char *errorString = "";

/*-------------------------------------------------------------------------*/

static Plugin methods = 
{
    MP3Shutdown,
    MP3GetVersion,
    MP3GetName,
    MP3GetNumFormats,
    MP3GetFormat,
    MP3GetError,
    MP3ReadMetadata,
    MP3WriteMetadata,
    MP3DecodeStart,
    MP3DecodeInfo,
    MP3DecodeRead,
    MP3DecodeEnd
};

/*-------------------------------------------------------------------------*/

Plugin *initPlugin(void)
{
    return &methods;
}

static void MP3Shutdown(void)
{
    if (strlen(errorString))
       free(errorString);
}

static const char *MP3GetVersion(void)
{
    return PLUGIN_VERSION;
}

static const char *MP3GetName(void)
{
    return PLUGIN_NAME;
}

static int MP3GetNumFormats(void)
{
    return 1;
}

static int MP3GetFormat(int i, char ext[TP_EXTENSION_LEN],
                        char desc[TP_PLUGIN_DESC_LEN], int *functions)
{
    if (i > 0)
        return 0;

    strcpy(ext, ".mp3");
    strcpy(desc, "MP3 Audio format");
    *functions = TP_PLUGIN_FUNCTION_DECODE | TP_PLUGIN_FUNCTION_METADATA;

    return 1;
}

static const char *MP3GetError(void)
{
    return errorString;
}

static int MP3ReadMetadata(metadata_t *mdata, const char *fileName, int flags)
{
    Metadata data;
    ID3      id3(true);

    if (id3.read(fileName, data))
    {
        data.writeToC(mdata);
        return 1;
    }
    return 0;
}

static int MP3WriteMetadata(const metadata_t *mdata, const char *fileName, int flags)
{
    Metadata data;
    ID3      id3((flags | PLUGIN_FLAGS_WRITE_ID3V1) != 0);

    data.readFromC(mdata);
    return id3.write(fileName, data, (flags | TP_PLUGIN_FLAGS_GENERAL_CLEAR_TAGS) != 0);
}

static void *MP3DecodeStart(const char *file)
{
    return (void *)new MP3Decode(file);
}

static int MP3DecodeInfo(void *decode,
                         unsigned long *duration,
                         unsigned int *samplesPerSecond,
                         unsigned int *bitsPerSample,
                         unsigned int *channels)
{
    return ((MP3Decode *)decode)->getInfo(*duration, *samplesPerSecond, *bitsPerSample, *channels); 
}

static int MP3DecodeRead(void *decode, char *data, int maxBytes)
{
    return ((MP3Decode *)decode)->read(data, maxBytes);
}

static void MP3DecodeEnd(void *decode)
{
    delete (MP3Decode *)decode;
}

} // extern "C"
